from collections.abc import Sized
from typing import cast
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import numpy.typing as npt
from scipy.sparse import coo_matrix

from ._typing_utils import Float
from ._typing_utils import Int
from ._typing_utils import IntArray
from ._typing_utils import Matrix
from ._typing_utils import NumArray


IndexType = Union[Sequence[Int], IntArray]
DataType = Union[Sequence[Union[Int, Float]], NumArray]


class coo_matrix_builder:  # noqa: N801
    """store data to build scipy.sparce.coo_matrix."""

    def __init__(
        self,
        shape: Tuple[Int, Int],
        row: Optional[IndexType] = None,
        col: Optional[IndexType] = None,
        data: Optional[DataType] = None,
        *,
        dtype: npt.DTypeLike = np.float64,
        index_dtype: npt.DTypeLike = np.int64,
    ) -> None:
        """Initialize the object.

        Parameters
        ----------
        n_row : Int
            The row count.
        n_col : Int
            The column count.
        row : Optional[Sequence[Int]], optional
            The row values. If None, an empty array will be created.
        col : Optional[Sequence[Int]], optional
            The column values. If None, an empty array will be created.
        data : Optional[Sequence[Union[Int, Float]]], optional
            The data values. If None, an empty array will be created.
        dtype : npt.DTypeLike, optional
            The dtype for the values, by default np.float64
        index_dtype : npt.DTypeLike, optional
            The dtype for the index, by default np.int64
        """
        assert len(shape) == 2
        if row is None:
            row = []
        if col is None:
            col = []
        if data is None:
            data = []
        self.n_row = shape[0]
        self.n_col = shape[1]
        assert len(row) == len(col)
        assert len(row) == len(data)
        self.row = np.array(row, dtype=index_dtype)
        self.col = np.array(col, dtype=index_dtype)
        self.data = np.array(data, dtype=dtype)
        self.shape = (self.n_row, self.n_col)
        self.dtype = dtype
        self.index_dtype = index_dtype

    def append(
        self,
        row: Union[Int, IndexType],
        col: Union[Int, IndexType],
        data: Union[Int, Float, DataType],
    ) -> None:
        """Append data from row, column index and data values.

        Parameters
        ----------
        row : Union[Int, IndexType]
            The row values.
        col : Union[Int, IndexType]
            The column values.
        data : Union[Int, Float, DataType]
            The data values.
        """
        if isinstance(row, Sized) or isinstance(col, Sized) or isinstance(data, Sized):
            count = None
            if isinstance(row, Sized):
                # dirty hack but will be solved in Python 3.10
                # https://stackoverflow.com/questions/65912706/pattern-matching-over-nested-union-types-in-python # noqa :
                row2 = np.array(cast(IndexType, row))
                count = len(row2)
                if isinstance(col, Sized):
                    col2 = np.array(cast(IndexType, col))
                    assert len(col2) == count
                else:
                    col2 = np.ones(count, dtype=self.index_dtype) * col
            else:
                assert isinstance(col, Sized)
                col2 = np.array(cast(IndexType, col))
                count = len(col2)
                row2 = np.ones(count, dtype=self.index_dtype) * row

            if isinstance(data, Sized):
                data2 = np.array(cast(DataType, data))
                assert len(data2) == count
            else:
                data2 = np.ones(count, dtype=self.dtype) * data
        else:
            row2 = np.array([cast(Int, row)])
            col2 = np.array([cast(Int, col)])
            data2 = np.array([cast(Union[Int, Float], data)])
            count = 1

        if count == 0:
            return
        assert len(row2) == count
        assert len(col2) == count
        assert len(data2) == count
        self.row = np.concatenate([self.row, row2], dtype=self.index_dtype)
        self.col = np.concatenate([self.col, col2], dtype=self.index_dtype)
        self.data = np.concatenate([self.data, data2], dtype=self.dtype)

    def append_matrix(
        self,
        matrix: Union["coo_matrix_builder", Matrix],
        shift: Tuple[Int, Int] = (0, 0),
    ) -> None:
        """Append data from another coo_matrix_builder.

        Parameters
        ----------
        matrix : coo_matrix_builder or Matrix
            the matrix to append
        shift : Tuple[Int,Int]
            the shift for row and column
        """
        assert len(shift)
        if not isinstance(matrix, coo_matrix_builder):
            matrix2 = coo_matrix(matrix)
        else:
            matrix2 = matrix
        self.append(matrix2.row + shift[0], matrix2.col + shift[1], matrix2.data)

    def to_coo_matrix(self) -> coo_matrix:
        """Generate `coo_matrix`.

        Returns
        -------
        matrix : coo_matrix
            the generated coo_matrix
        """
        return coo_matrix(
            (self.data, (self.row, self.col)), shape=(self.n_row, self.n_col)
        )

    def __setitem__(
        self, index: Tuple[Union[Int, IndexType], Union[Int, IndexType]], value
    ):
        """Syntax sugar for append.

        Parameters
        ----------
        index : Union[Tuple[Int], Tuple[Sequence[Int]]]
            The index to append values. The length must be 2 (row,col).
        value : [type]
            The data values to append.
        """
        assert len(index) == 2
        self.append(index[0], index[1], value)

    @property
    def T(self) -> "coo_matrix_builder":  # noqa : N802
        """Transpose matrix.

        Returns
        -------
        builder : coo_matrix_builder
            the transposed builder
        """
        return coo_matrix_builder(self.shape[::-1], self.col, self.row, self.data)

    def size(self):
        """Get current array size.

        Returns
        -------
        size : int
            the array size for data
        """
        assert len(self.row) == len(self.col)
        assert len(self.row) == len(self.data)
        return len(self.data)



from typing import Optional
from typing import Union

import numpy as np
from scipy.sparse import coo_matrix

from ._coo_matrix_builder import coo_matrix_builder
from ._typing_utils import Float
from ._typing_utils import Matrix


def build_frame_cost_matrix(
    dist_matrix: coo_matrix_builder,
    *,
    track_start_cost: Optional[Float],
    track_end_cost: Optional[Float],
) -> coo_matrix:
    """Build sparce array for frame-linking cost matrix.

    Parameters
    ----------
    dist_matrix : Matrix or `_utils.coo_matrix_builder`
        The distance matrix for points at time t and t+1.
    track_start_cost : Float, optional
        The cost for starting the track (b in Jaqaman et al 2008 NMeth)
    track_end_cost : Float, optional
        The cost for ending the track (d in Jaqaman et al 2008 NMeth)

    Returns
    -------
    cost_matrix : FloatArray
        the cost matrix for frame linking
    """
    M = dist_matrix.shape[0]
    N = dist_matrix.shape[1]

    C = coo_matrix_builder((M + N, N + M), dtype=np.float32)
    C.append_matrix(dist_matrix)

    if track_start_cost is None:
        if len(C.data) > 0:
            track_start_cost = np.max(C.data) * 1.05
        else:
            track_start_cost = 1.05
    if track_end_cost is None:
        if len(C.data) > 0:
            track_end_cost = np.max(C.data) * 1.05
        else:
            track_end_cost = 1.05

    C[np.arange(M, M + N), np.arange(N)] = np.ones(N) * track_end_cost
    C[np.arange(M), np.arange(N, N + M)] = np.ones(M) * track_start_cost
    min_val = np.min(C.data) if len(C.data) > 0 else 0
    C[dist_matrix.col + M, dist_matrix.row + N] = min_val

    return C.to_coo_matrix()


def build_segment_cost_matrix(
    gap_closing_dist_matrix: Union[coo_matrix_builder, Matrix],
    splitting_dist_matrix: Union[coo_matrix_builder, Matrix],
    merging_dist_matrix: Union[coo_matrix_builder, Matrix],
    track_start_cost: Optional[Float],
    track_end_cost: Optional[Float],
    no_splitting_cost: Optional[Float],
    no_merging_cost: Optional[Float],
    alternative_cost_factor: Float = 1.05,
    alternative_cost_percentile: Float = 90,
    alternative_cost_percentile_interpolation: str = "lower",
) -> Optional[coo_matrix]:
    """Build sparce array for segment-linking cost matrix.

    Parameters
    ----------
    gap_closing_dist_matrix : coo_matrix_builder or Matrix
        The distance matrix for closing gaps between segment i and j.
    splitting_dist_matrix : coo_matrix_builder or Matrix
        The distance matrix for splitting between segment i and time/index j
    merging_dist_matrix : coo_matrix_builder or Matrix
        The distance matrix for merging between segment i and time/index j
    track_start_cost : Float, optional
        The cost for starting the track (b in Jaqaman et al 2008 NMeth)
    track_end_cost : Float, optional
        The cost for ending the track (d in Jaqaman et al 2008 NMeth)
    no_splitting_cost : Float, optional
        The cost to reject splitting (d' in Jaqaman et al 2008 NMeth)
    no_merging_cost : Float, optional
        The cost to reject merging (b' in Jaqaman et al 2008 NMeth)
    alternative_cost_factor: Float
        The factor to calculate the alternative costs, by default 1.05.
    alternative_cost_percentile: Float
        The percentile to calculate the alternative costs, by default 90.
    alternative_cost_percentile_interpolation: str
        The percentile interpolation to calculate the alternative costs,
        by default "lower".
        See `numpy.percentile` for allowed values.

    Returns
    -------
    cost_matrix : Optional[coo_matrix]
        the cost matrix for frame linking, None if not appropriate
    """
    M = gap_closing_dist_matrix.shape[0]
    assert gap_closing_dist_matrix.shape[1] == M
    assert splitting_dist_matrix.shape[0] == M
    assert merging_dist_matrix.shape[0] == M
    N1 = splitting_dist_matrix.shape[1]
    N2 = merging_dist_matrix.shape[1]

    S = 2 * M + N1 + N2

    C = coo_matrix_builder((S, S), dtype=np.float32)

    C.append_matrix(gap_closing_dist_matrix)
    C.append_matrix(splitting_dist_matrix.T, shift=(M, 0))
    C.append_matrix(merging_dist_matrix, shift=(0, M))

    upper_left_size = C.size()
    if upper_left_size == 0:
        return None

    # Note:
    # Though the way of assigning track_start_cost, track_end_cost, no_splitting_cost, no_merging_cost # noqa :
    # and min_val is similar to that of TrackMate (link1, link2), GPL3 of TrackMate does not apply. (See link3 for license discussion.) # noqa :
    #   link1 https://github.com/fiji/TrackMate/blob/5a97426586b3c592c986c57aa1a09bab9d21419c/src/main/java/fiji/plugin/trackmate/tracking/sparselap/costmatrix/DefaultCostMatrixCreator.java#L186 # noqa :
    #         https://github.com/fiji/TrackMate/blob/5a97426586b3c592c986c57aa1a09bab9d21419c/src/main/java/fiji/plugin/trackmate/tracking/sparselap/costmatrix/JaqamanSegmentCostMatrixCreator.java # noqa:
    #         https://github.com/fiji/TrackMate/blob/5a97426586b3c592c986c57aa1a09bab9d21419c/src/main/java/fiji/plugin/trackmate/tracking/sparselap/SparseLAPSegmentTracker.java#L148 # noqa:
    #   link2 (default parameters for alternative_cost_percentile, alternative_cost_factor) # noqa :
    #         https://github.com/fiji/TrackMate/blob/5a97426586b3c592c986c57aa1a09bab9d21419c/src/main/java/fiji/plugin/trackmate/tracking/TrackerKeys.java # noqa :
    #   link3 https://forum.image.sc/t/linear-assignment-problem-based-tracking-package-in-python/57793 # noqa :
    #         https://web.archive.org/web/20210921134401/https://forum.image.sc/t/linear-assignment-problem-based-tracking-package-in-python/57793 # noqa :

    if (
        track_start_cost is None
        or track_end_cost is None
        or no_splitting_cost is None
        or no_merging_cost is None
    ):
        alternative_cost = (
            np.percentile(
                # XXX seems numpy / mypy is over-strict here. Will fix later.
                C.data,  # type: ignore
                alternative_cost_percentile,
                method=alternative_cost_percentile_interpolation,
            )
            * alternative_cost_factor
        )
        if track_start_cost is None:
            track_start_cost = alternative_cost
        if track_end_cost is None:
            track_end_cost = alternative_cost
        if no_splitting_cost is None:
            no_splitting_cost = alternative_cost
        if no_merging_cost is None:
            no_merging_cost = alternative_cost

    C[np.arange(M + N1, 2 * M + N1), np.arange(M)] = np.ones(M) * track_start_cost
    C[np.arange(2 * M + N1, S), np.arange(M, M + N2)] = np.ones(N2) * no_merging_cost
    C[np.arange(M), np.arange(M + N2, 2 * M + N2)] = np.ones(M) * track_end_cost
    C[np.arange(M, M + N1), np.arange(2 * M + N2, S)] = np.ones(N1) * no_splitting_cost
    min_val = np.min(
        np.array([track_start_cost, track_end_cost, no_splitting_cost, no_merging_cost])
    )

    upper_left_rows = C.row[:upper_left_size]
    upper_left_cols = C.col[:upper_left_size]
    C[upper_left_cols + M + N1, upper_left_rows + M + N2] = min_val

    return C.to_coo_matrix()



"""Main module for tracking."""
import logging
from abc import ABC
from abc import abstractmethod
from inspect import Parameter
from inspect import signature
from typing import Callable
from typing import cast
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Union

# https://stackoverflow.com/questions/59037244/mypy-incompatible-import-error-for-conditional-imports # noqa :
if TYPE_CHECKING:
    from typing import Literal
else:
    try:
        from typing import Literal
    except ImportError:
        from typing_extensions import Literal

import networkx as nx
import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist
from scipy.sparse import coo_matrix
from pydantic import BaseModel, Field, Extra


from ._cost_matrix import build_frame_cost_matrix, build_segment_cost_matrix
from ._optimization import lap_optimization
from ._typing_utils import FloatArray
from ._typing_utils import Int
from ._coo_matrix_builder import coo_matrix_builder

logger = logging.getLogger(__name__)


def _get_segment_df(coords, track_tree):
    # linking between tracks
    segments = list(nx.connected_components(track_tree))
    first_nodes = np.array(
        list(map(lambda segment: min(segment, key=lambda node: node[0]), segments))
    )
    last_nodes = np.array(
        list(map(lambda segment: max(segment, key=lambda node: node[0]), segments))
    )
    segments_df = pd.DataFrame(
        {
            "segment": segments,
            "first_frame": first_nodes[:, 0],
            "first_index": first_nodes[:, 1],
            "last_frame": last_nodes[:, 0],
            "last_index": last_nodes[:, 1],
        }
    ).reset_index()

    for prefix in ["first", "last"]:
        segments_df[f"{prefix}_frame_coords"] = segments_df.apply(
            lambda row: coords[row[f"{prefix}_frame"]][row[f"{prefix}_index"]],
            axis=1,
        )
    return segments_df


def _get_segment_end_connecting_matrix(
    segments_df, max_frame_count, dist_metric, cost_cutoff
):
    if cost_cutoff:

        def to_gap_closing_candidates(row):
            target_coord = row["last_frame_coords"]
            frame_diff = segments_df["first_frame"] - row["last_frame"]
            indices = (1 <= frame_diff) & (frame_diff <= max_frame_count)
            df = segments_df[indices]
            # note: can use KDTree if metric is distance,
            # but might not be appropriate for general metrics
            # https://stackoverflow.com/questions/35459306/find-points-within-cutoff-distance-of-other-points-with-scipy # noqa
            # TrackMate also uses this (trivial) implementation.
            if len(df) > 0:
                target_dist_matrix = cdist(
                    [target_coord],
                    np.stack(df["first_frame_coords"].values),
                    metric=dist_metric,
                )
                assert target_dist_matrix.shape[0] == 1
                indices2 = np.where(target_dist_matrix[0] < cost_cutoff)[0]
                return (
                    df.index[indices2].values,
                    target_dist_matrix[0][indices2],
                )
            else:
                return [], []

        segments_df["gap_closing_candidates"] = segments_df.apply(
            to_gap_closing_candidates, axis=1
        )
    else:
        segments_df["gap_closing_candidates"] = [([], [])] * len(segments_df)

    N_segments = len(segments_df)
    gap_closing_dist_matrix = coo_matrix_builder(
        (N_segments, N_segments), dtype=np.float32
    )
    for ind, row in segments_df.iterrows():
        candidate_inds = row["gap_closing_candidates"][0]
        candidate_costs = row["gap_closing_candidates"][1]
        # row ... track end, col ... track start
        gap_closing_dist_matrix[(int(cast(int, ind)), candidate_inds)] = candidate_costs

    return segments_df, gap_closing_dist_matrix


def _get_splitting_merging_candidates(
    segments_df,
    coords,
    cutoff,
    prefix,
    dist_metric,
):
    if cutoff:

        def to_candidates(row):
            target_coord = row[f"{prefix}_frame_coords"]
            frame = row[f"{prefix}_frame"] + (-1 if prefix == "first" else 1)
            # note: can use KDTree if metric is distance,
            # but might not be appropriate for general metrics
            # https://stackoverflow.com/questions/35459306/find-points-within-cutoff-distance-of-other-points-with-scipy # noqa
            if frame < 0 or len(coords) <= frame:
                return [], []
            target_dist_matrix = cdist(
                [target_coord], coords[frame], metric=dist_metric
            )
            assert target_dist_matrix.shape[0] == 1
            indices = np.where(target_dist_matrix[0] < cutoff)[0]
            return [(frame, index) for index in indices], target_dist_matrix[0][indices]

        segments_df[f"{prefix}_candidates"] = segments_df.apply(to_candidates, axis=1)
    else:
        segments_df[f"{prefix}_candidates"] = [([], [])] * len(segments_df)

    middle_point_candidates = np.unique(
        sum(
            segments_df[f"{prefix}_candidates"].apply(lambda x: list(x[0])),
            [],
        ),
        axis=0,
    )

    N_segments = len(segments_df)
    N_middle = len(middle_point_candidates)
    dist_matrix = coo_matrix_builder((N_segments, N_middle), dtype=np.float32)

    middle_point_candidates_dict = {
        tuple(val): i for i, val in enumerate(middle_point_candidates)
    }
    for ind, row in segments_df.iterrows():
        candidate_frame_indices = row[f"{prefix}_candidates"][0]
        candidate_inds = [
            middle_point_candidates_dict[tuple(fi)] for fi in candidate_frame_indices
        ]
        candidate_costs = row[f"{prefix}_candidates"][1]
        dist_matrix[(int(cast(Int, ind)), candidate_inds)] = candidate_costs

    return segments_df, dist_matrix, middle_point_candidates


def _remove_no_split_merge_links(track_tree, segment_connected_edges):
    for edge in segment_connected_edges:
        assert len(edge) == 2
        younger, elder = edge
        # if the edge is involved with branching or merging, do not remove the edge
        if (
            sum([int(node[0] > younger[0]) for node in track_tree.neighbors(younger)])
            > 1
        ):
            continue
        if sum([int(node[0] < elder[0]) for node in track_tree.neighbors(elder)]) > 1:
            continue
        track_tree.remove_edge(*edge)
    return track_tree


class LapTrackBase(BaseModel, ABC, extra=Extra.forbid):
    track_dist_metric: Union[str, Callable] = Field(
        "sqeuclidean",
        description="The metric for calculating track linking cost, "
        + "See documentation for `scipy.spatial.distance.cdist` for accepted values.",
    )
    splitting_dist_metric: Union[str, Callable] = Field(
        "sqeuclidean",
        description="The metric for calculating splitting cost."
        + "See `track_dist_metric`",
    )
    merging_dist_metric: Union[str, Callable] = Field(
        "sqeuclidean",
        description="The metric for calculating merging cost."
        + "See `track_dist_metric`",
    )

    alternative_cost_factor: float = Field(
        1.05,
        description="The factor to calculate the alternative costs"
        + "(b,d,b',d' in Jaqaman et al 2008 NMeth)",
    )
    alternative_cost_percentile: float = Field(
        90,
        description="The percentile to calculate the alternative costs"
        + "(b,d,b',d' in Jaqaman et al 2008 NMeth)",
    )
    alternative_cost_percentile_interpolation: str = Field(
        "lower",
        description="The percentile interpolation to calculate the alternative costs"
        + "(b,d,b',d' in Jaqaman et al 2008 NMeth)."
        + "See `numpy.percentile` for accepted values.",
    )

    track_cost_cutoff: float = Field(
        15**2,
        description="The cost cutoff for the connected points in the track."
        + "For default cases with `dist_metric='sqeuclidean'`,"
        + "this value should be squared maximum distance.",
    )
    track_start_cost: Optional[float] = Field(
        None,  # b in Jaqaman et al 2008 NMeth.
        description="The cost for starting the track (b in Jaqaman et al 2008 NMeth),"
        + "if None, automatically estimated",
    )
    track_end_cost: Optional[float] = Field(
        None,  # b in Jaqaman et al 2008 NMeth.
        description="The cost for ending the track (b in Jaqaman et al 2008 NMeth),"
        + "if None, automatically estimated",
    )
    segment_start_cost: Optional[float] = Field(
        None,  # b in Jaqaman et al 2008 NMeth for segment connection
        description="The cost for starting the segment (b in Jaqaman et al 2008 NMeth),"
        + "if None, automatically estimated",
    )
    segment_end_cost: Optional[float] = Field(
        None,  # b in Jaqaman et al 2008 NMeth for segment connection
        description="The cost for ending the segment (b in Jaqaman et al 2008 NMeth),"
        + "if None, automatically estimated",
    )

    gap_closing_cost_cutoff: Union[Literal[False], float] = Field(
        15**2,
        description="The cost cutoff for gap closing."
        + "For default cases with `dist_metric='sqeuclidean'`,"
        + "this value should be squared maximum distance."
        + "If False, no gap closing is allowed.",
    )

    gap_closing_max_frame_count: int = Field(
        2, description="The maximum frame gaps, by default 2."
    )

    splitting_cost_cutoff: Union[Literal[False], float] = Field(
        False,
        description="The cost cutoff for splitting."
        + "See `gap_closing_cost_cutoff`."
        + "If False, no splitting is allowed.",
    )

    no_splitting_cost: Optional[float] = Field(
        None,  # d' in Jaqaman et al 2008 NMeth.
        description="The cost to reject splitting, if None, automatically estimated.",
    )

    merging_cost_cutoff: Union[Literal[False], float] = Field(
        False,
        description="The cost cutoff for merging."
        + "See `gap_closing_cost_cutoff`."
        + "If False, no merging is allowed.",
    )

    no_merging_cost: Optional[float] = Field(
        None,  # d' in Jaqaman et al 2008 NMeth.
        description="The cost to reject merging, if None, automatically estimated.",
    )

    def _link_frames(self, coords) -> nx.Graph:
        """Link particles between frames according to the cost function

        Args:
            coords (List[np.ndarray]): the input coordinates

        Returns:
            nx.Graph: the resulted tree
        """
        # initialize tree
        track_tree = nx.Graph()
        for frame, coord in enumerate(coords):
            for j in range(coord.shape[0]):
                track_tree.add_node((frame, j))

        # linking between frames
        for frame, (coord1, coord2) in enumerate(zip(coords[:-1], coords[1:])):
            dist_matrix = cdist(coord1, coord2, metric=self.track_dist_metric)
            ind = np.where(dist_matrix < self.track_cost_cutoff)
            dist_matrix = coo_matrix_builder(
                dist_matrix.shape,
                row=ind[0],
                col=ind[1],
                data=dist_matrix[(*ind,)],
                dtype=dist_matrix.dtype,
            )
            cost_matrix = build_frame_cost_matrix(
                dist_matrix,
                track_start_cost=self.track_start_cost,
                track_end_cost=self.track_end_cost,
            )
            _, xs, _ = lap_optimization(cost_matrix)

            count1 = dist_matrix.shape[0]
            count2 = dist_matrix.shape[1]
            connections = [(i, xs[i]) for i in range(count1) if xs[i] < count2]
            # track_start=[i for i in range(count1) if xs[i]>count2]
            # track_end=[i for i in range(count2) if ys[i]>count1]
            for connection in connections:
                track_tree.add_edge((frame, connection[0]), (frame + 1, connection[1]))
        return track_tree

    def _get_gap_closing_matrix(self, segments_df):
        return _get_segment_end_connecting_matrix(
            segments_df,
            self.gap_closing_max_frame_count,
            self.track_dist_metric,
            self.gap_closing_cost_cutoff,
        )

    def _link_gap_split_merge_from_matrix(
        self,
        segments_df,
        track_tree,
        gap_closing_dist_matrix,
        splitting_dist_matrix,
        merging_dist_matrix,
        splitting_all_candidates,
        merging_all_candidates,
    ):
        cost_matrix = build_segment_cost_matrix(
            gap_closing_dist_matrix,
            splitting_dist_matrix,
            merging_dist_matrix,
            self.segment_start_cost,
            self.segment_end_cost,
            self.no_splitting_cost,
            self.no_merging_cost,
            self.alternative_cost_factor,
            self.alternative_cost_percentile,
            self.alternative_cost_percentile_interpolation,
        )

        if not cost_matrix is None:
            _, xs, ys = lap_optimization(cost_matrix)

            M = gap_closing_dist_matrix.shape[0]
            N1 = splitting_dist_matrix.shape[1]
            N2 = merging_dist_matrix.shape[1]

            for ind, row in segments_df.iterrows():
                col_ind = xs[ind]
                first_frame_index = (row["first_frame"], row["first_index"])
                last_frame_index = (row["last_frame"], row["last_index"])
                if col_ind < M:
                    target_frame_index = tuple(
                        segments_df.loc[col_ind, ["first_frame", "first_index"]]
                    )
                    track_tree.add_edge(last_frame_index, target_frame_index)
                elif col_ind < M + N2:
                    track_tree.add_edge(
                        last_frame_index,
                        tuple(merging_all_candidates[col_ind - M]),
                    )

                row_ind = ys[ind]
                if M <= row_ind and row_ind < M + N1:
                    track_tree.add_edge(
                        first_frame_index,
                        tuple(splitting_all_candidates[row_ind - M]),
                    )

        return track_tree

    @abstractmethod
    def _predict_gap_split_merge(self, coords, track_tree):
        ...

    def predict(self, coords) -> nx.Graph:
        """Predict the tracking graph from coordinates

        Args:
            coords : Sequence[FloatArray]
                The list of coordinates of point for each frame.
                The array index means (sample, dimension).


        Raises:
            ValueError: raised for invalid coordinate formats.

        Returns:
            nx.Graph: The graph for the tracks, whose nodes are (frame, index).
        """

        if any(list(map(lambda coord: coord.ndim != 2, coords))):
            raise ValueError("the elements in coords must be 2-dim.")
        coord_dim = coords[0].shape[1]
        if any(list(map(lambda coord: coord.shape[1] != coord_dim, coords))):
            raise ValueError("the second dimension in coords must have the same size")

        ####### Particle-particle tracking #######
        track_tree = self._link_frames(coords)
        track_tree = self._predict_gap_split_merge(coords, track_tree)
        return track_tree


class LapTrack(LapTrackBase):
    def _predict_gap_split_merge(self, coords, track_tree):
        """one-step fitting, as TrackMate and K. Jaqaman et al., Nat Methods 5, 695 (2008).

        Args:
            coords : Sequence[FloatArray]
                The list of coordinates of point for each frame.
                The array index means (sample, dimension).
            track_tree : nx.Graph
                the track tree

        Returns:
            track_tree : nx.Graph
                the updated track tree
        """
        if (
            self.gap_closing_cost_cutoff
            or self.splitting_cost_cutoff
            or self.merging_cost_cutoff
        ):
            segments_df = _get_segment_df(coords, track_tree)

            # compute candidate for gap closing
            segments_df, gap_closing_dist_matrix = self._get_gap_closing_matrix(
                segments_df
            )

            middle_points: Dict = {}
            dist_matrices: Dict = {}

            # compute candidate for splitting and merging
            for prefix, cutoff, dist_metric in zip(
                ["first", "last"],
                [self.splitting_cost_cutoff, self.merging_cost_cutoff],
                [self.splitting_dist_metric, self.merging_dist_metric],
            ):
                (
                    segments_df,
                    dist_matrices[prefix],
                    middle_points[prefix],
                ) = _get_splitting_merging_candidates(
                    segments_df, coords, cutoff, prefix, dist_metric
                )

            splitting_dist_matrix = dist_matrices["first"]
            merging_dist_matrix = dist_matrices["last"]
            splitting_all_candidates = middle_points["first"]
            merging_all_candidates = middle_points["last"]
            track_tree = self._link_gap_split_merge_from_matrix(
                segments_df,
                track_tree,
                gap_closing_dist_matrix,
                splitting_dist_matrix,
                merging_dist_matrix,
                splitting_all_candidates,
                merging_all_candidates,
            )

        return track_tree


class LapTrackMulti(LapTrackBase):
    segment_connecting_metric: Union[str, Callable] = Field(
        "sqeuclidean",
        description="The metric for calculating cost to connect segment ends."
        + "See `track_dist_metric`.",
    )
    segment_connecting_cost_cutoff: float = Field(
        15**2,
        description="The cost cutoff for splitting." + "See `gap_closing_cost_cutoff`.",
    )

    remove_no_split_merge_links: bool = Field(
        False,
        description="if True, remove segment connections if splitting did not happen.",
    )

    def _get_segment_connecting_matrix(self, segments_df):
        return _get_segment_end_connecting_matrix(
            segments_df,
            1,  # only arrow 1-frame difference
            self.segment_connecting_metric,
            self.segment_connecting_cost_cutoff,
        )

    def _predict_gap_split_merge(self, coords, track_tree):
        # "multi-step" type of fitting (Y. T. Fukai (2022))
        segments_df = _get_segment_df(coords, track_tree)

        ###### gap closing step ######
        ###### split - merge step 1 ######

        get_matrix_fns = {
            "gap_closing": self._get_gap_closing_matrix,
            "segment_connecting": self._get_segment_connecting_matrix,
        }

        segment_connected_edges = []
        for mode, get_matrix_fn in get_matrix_fns.items():
            segments_df, gap_closing_dist_matrix = get_matrix_fn(segments_df)
            cost_matrix = build_frame_cost_matrix(
                gap_closing_dist_matrix,
                track_start_cost=self.segment_start_cost,
                track_end_cost=self.segment_end_cost,
            )
            _, xs, _ = lap_optimization(cost_matrix)

            nrow = gap_closing_dist_matrix.shape[0]
            ncol = gap_closing_dist_matrix.shape[1]
            connections = [(i, xs[i]) for i in range(nrow) if xs[i] < ncol]
            for connection in connections:
                # connection ... connection segments_df.iloc[i] -> segments_df.iloc[xs[i]]
                node_from = tuple(
                    segments_df.loc[connection[0], ["last_frame", "last_index"]]
                )
                node_to = tuple(
                    segments_df.loc[connection[1], ["first_frame", "first_index"]]
                )
                track_tree.add_edge(node_from, node_to)
                if mode == "segment_connecting":
                    segment_connected_edges.append((node_from, node_to))

            # regenerate segments after closing gaps
            segments_df = _get_segment_df(coords, track_tree)

        ###### split - merge step 2 ######
        middle_points: Dict = {}
        dist_matrices: Dict = {}
        for prefix, cutoff, dist_metric in zip(
            ["first", "last"],
            [self.splitting_cost_cutoff, self.merging_cost_cutoff],
            [self.splitting_dist_metric, self.merging_dist_metric],
        ):
            dist_metric_argnums = None
            if callable(dist_metric):
                try:
                    s = signature(dist_metric)
                    dist_metric_argnums = len(
                        [
                            0
                            for p in s.parameters.values()
                            if p.kind == Parameter.POSITIONAL_OR_KEYWORD
                            or p.kind == Parameter.POSITIONAL_ONLY
                        ]
                    )
                except TypeError:
                    pass
            if callable(dist_metric) and dist_metric_argnums >= 3:
                logger.info("using callable dist_metric with more than 2 parameters")
                # the dist_metric function is assumed to take
                # (coordinate1, coordinate2, coordinate_sibring, connected by segment_connecting step)
                segment_connected_nodes = [
                    e[0 if prefix == "first" else 1] for e in segment_connected_edges
                ]  # find nodes connected by "segment_connect" steps
                _coords = [
                    [(*c, frame, ind) for ind, c in enumerate(coord_frame)]
                    for frame, coord_frame in enumerate(coords)
                ]
                assert np.all(c.shape[1] == _coords[0].shape[1] for c in _coords)

                # _coords ... (coordinate, frame, if connected by segment_connecting step)
                def _dist_metric(c1, c2):
                    *_c1, frame1, ind1 = c1
                    *_c2, frame2, ind2 = c2
                    # for splitting case, check the yonger one
                    if not frame1 < frame2:
                        # swap frame1 and 2; always assume coordinate 1 is first
                        tmp = _c1, frame1, ind1
                        _c1, frame1, ind1 = _c2, frame2, ind2
                        _c2, frame2, ind2 = tmp
                    check_node = (frame1, ind1) if prefix == "first" else (frame2, ind2)
                    if dist_metric_argnums == 3:
                        return dist_metric(
                            np.array(_c1),
                            np.array(_c2),
                            check_node in segment_connected_nodes,
                        )
                    else:
                        if prefix == "first":
                            # splitting sibring candidate
                            candidates = [
                                (frame, ind)
                                for (frame, ind) in track_tree.neighbors((frame1, ind1))
                                if frame > frame1
                            ]
                        else:
                            # merging sibring candidate
                            candidates = [
                                (frame, ind)
                                for (frame, ind) in track_tree.neighbors((frame2, ind2))
                                if frame < frame2
                            ]

                        if len(candidates) == 0:
                            c_sib = None
                        else:
                            assert len(candidates) == 1
                            c_sib = candidates[0]
                        return dist_metric(
                            np.array(_c1),
                            np.array(_c2),
                            np.array(coords[c_sib[0]][c_sib[1]]) if c_sib else None,
                            check_node in segment_connected_nodes,
                        )

                segments_df[f"{prefix}_frame_coords"] = segments_df.apply(
                    lambda row: (
                        *row[f"{prefix}_frame_coords"],
                        int(row[f"{prefix}_frame"]),
                        int(row[f"{prefix}_index"]),
                    ),
                    axis=1,
                )

            else:
                logger.info("using callable dist_metric with 2 parameters")
                _coords = coords
                _dist_metric = dist_metric

            (
                segments_df,
                dist_matrices[prefix],
                middle_points[prefix],
            ) = _get_splitting_merging_candidates(
                segments_df, _coords, cutoff, prefix, _dist_metric
            )

        splitting_dist_matrix = dist_matrices["first"]
        merging_dist_matrix = dist_matrices["last"]
        splitting_all_candidates = middle_points["first"]
        merging_all_candidates = middle_points["last"]
        N_segments = len(segments_df)
        track_tree = self._link_gap_split_merge_from_matrix(
            segments_df,
            track_tree,
            coo_matrix((N_segments, N_segments), dtype=np.float32),  # no gap closing
            splitting_dist_matrix,
            merging_dist_matrix,
            splitting_all_candidates,
            merging_all_candidates,
        )

        ###### remove segment connections if not associated with split / merge ######

        if self.remove_no_split_merge_links:
            track_tree = _remove_no_split_merge_links(
                track_tree.copy(), segment_connected_edges
            )
        return track_tree


def laptrack(coords: Sequence[FloatArray], **kwargs) -> nx.Graph:
    """Track points by solving linear assignment problem.

    Parameters
    ----------
    coords : Sequence[FloatArray]
        The list of coordinates of point for each frame.
        The array index means (sample, dimension).

    **kwargs : dict
        Parameters for the LapTrack initalization

    Returns
    -------
    tracks : networkx.Graph
        The graph for the tracks, whose nodes are (frame, index).

    """
    lt = LapTrack(**kwargs)
    return lt.predict(coords)



from typing import Tuple

import lap
from scipy.sparse import csr_matrix

from ._typing_utils import FloatArray
from ._typing_utils import Int
from ._typing_utils import IntArray
from ._typing_utils import Matrix


def __to_lap_sparse(
    cost_matrix: Matrix,
) -> Tuple[Int, FloatArray, IntArray, IntArray]:
    """Convert data for lap.lapmod."""
    n = cost_matrix.shape[0]
    cost_matrix2 = csr_matrix(cost_matrix)
    assert cost_matrix2.has_sorted_indices
    return n, cost_matrix2.data, cost_matrix2.indptr, cost_matrix2.indices


def lap_optimization(cost_matrix: Matrix) -> Tuple[float, IntArray, IntArray]:
    """Solves the linear assignment problem for a sparse matrix.

    Parameters
    ----------
    cost_matrix : lil_matrix or coo_matrix
        the cost matrix in scipy.sparse.lil_matrix format

    Returns
    -------
    cost : float
        the final estimated cost
    xs : IntArray
        the indices such that assigned indices are (i,x[i]) (i=0 ... )
    ys : IntArray
        the indices such that assigned indices are (y[j],j) (j=0 ... )
    """
    return lap.lapmod(*__to_lap_sparse(cost_matrix))



from typing import Union

import numpy as np
import numpy.typing as npt
from scipy.sparse import coo_matrix
from scipy.sparse import lil_matrix

NumArray = npt.NDArray[Union[np.float_, np.int_]]
FloatArray = npt.NDArray[np.float_]
IntArray = npt.NDArray[np.int_]

Int = Union[int, np.int_]
Float = Union[float, np.float_]

Matrix = Union[FloatArray, coo_matrix, lil_matrix]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['laptrack']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'lap>=0.4.0,<0.5.0',
 'networkx>=2.6.1,<3.0.0',
 'numpy>=1.22.0,<2.0.0',
 'pandas>=1.3.1,<2.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'scipy>=1.7.0,<2.0.0',
 'typing-extensions>=3.10.0,<4.0.0']

entry_points = \
{'console_scripts': ['laptrack = laptrack.__main__:main']}

setup_kwargs = {
    'name': 'laptrack',
    'version': '0.1.7a3',
    'description': 'LapTrack',
    'long_description': "LapTrack\n========\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black| |Zenodo|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/laptrack.svg\n   :target: https://pypi.org/project/laptrack/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/laptrack.svg\n   :target: https://pypi.org/project/laptrack/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/laptrack\n   :target: https://pypi.org/project/laptrack\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/laptrack\n   :target: https://opensource.org/licenses/BSD-3-Clause\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/laptrack/latest.svg?label=Read%20the%20Docs\n   :target: https://laptrack.readthedocs.io/\n   :alt: Read the documentation at https://laptrack.readthedocs.io/\n.. |Tests| image:: https://github.com/yfukai/laptrack/workflows/Tests/badge.svg\n   :target: https://github.com/yfukai/laptrack/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/yfukai/laptrack/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/yfukai/laptrack\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n.. |Zenodo| image:: https://zenodo.org/badge/DOI/10.5281/zenodo.5519538.svg\n   :target: https://doi.org/10.5281/zenodo.5519538\n   :alt: Zenodo\n\nFeatures\n--------\n\nProvides a robust particle tracking algorithm using the Linear Assignment Problem, with various cost functions for linking.\n\nInstallation\n------------\n\nYou can install *LapTrack* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install laptrack\n\n\nUsage\n-----\n\nPlease see the Usage_ for details.\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `The 3-Clause BSD License`_,\n*LapTrack* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\n- This program implements a modified version of the algorithm in the `K. Jaqaman et al. (2008)`_.\n\n- Inspired by TrackMate_ a lot. See documentation_ for its detailed algorithm, the `2016 paper`_, and the `2021 paper`_.\n\n- This project was generated from `@cjolowicz`_'s `Hypermodern Python Cookiecutter`_ template.\n\n\nCitation\n--------\n\nIf you use this program for your research, please cite it and help us build more.\n\n.. code-block:: bib\n\n   @misc{laptrack,\n      author = {Yohsuke T. Fukai},\n      title = {laptrack},\n      year  = {2021},\n      url   = {https://doi.org/10.5281/zenodo.5519537},\n   }\n\n\n.. _K. Jaqaman et al. (2008): https://www.nature.com/articles/nmeth.1237\n.. _TrackMate: https://imagej.net/plugins/trackmate/\n.. _documentation: https://imagej.net/plugins/trackmate/algorithms\n.. _2016 paper: https://doi.org/10.1016/j.ymeth.2016.09.016\n.. _2021 paper: https://doi.org/10.1101/2021.09.03.458852\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _The 3-Clause BSD License: https://opensource.org/licenses/BSD-3-Clause\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/yfukai/laptrack/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://laptrack.readthedocs.io/en/latest/usage.html\n",
    'author': 'Yohsuke Fukai',
    'author_email': 'ysk@yfukai.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yfukai/laptrack',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<3.11',
}


setup(**setup_kwargs)



"""LapTrack: particle tracking by linear assignment problem."""

__author__ = """Yohsuke T. Fukai"""
__email__ = "ysk@yfukai.net"

from ._tracking import laptrack, LapTrack, LapTrackMulti
from . import data_conversion

__all__ = ["laptrack", "LapTrack", "LapTrackMulti", "data_conversion"]



from typing import List
from typing import Tuple

import networkx as nx
import numpy as np
import pandas as pd

from ._typing_utils import Int
from ._typing_utils import NumArray

IntTuple = Tuple[Int, Int]


def convert_dataframe_to_coords(
    df: pd.DataFrame,
    coordinate_cols: List[str],
    frame_col: str = "frame",
    validate_frame: bool = True,
) -> List[NumArray]:
    """Convert a track dataframe to a list of coordinates for input.

    Parameters
    ----------
    df : pd.DataFrame
        the input dataframe
    coordinate_cols : List[str]
        the list of columns to use for coordinates
    frame_col : str, optional
        The column name to use for the frame index. Defaults to "frame".
    validate_frame : bool, optional
        whether to validate the frame. Defaults to True.

    Returns
    -------
    coords : List[np.ndarray]
        the list of coordinates
    """

    grps = list(df.groupby(frame_col, sort=True))
    if validate_frame:
        assert np.array_equal(np.arange(df[frame_col].max() + 1), [g[0] for g in grps])
    coords = [grp[coordinate_cols].values for _frame, grp in grps]
    return coords


def convert_tree_to_dataframe(
    tree: nx.Graph,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Convert the track tree to dataframes

    Parameters
    ----------
    tree : nx.Graph
        The track tree, resulted from the traking

    Returns
    -------
    df : pd.DataFrame
        the track dataframe, with the following columns:
        - "frame" : the frame index
        - "index" : the coordinate index
        - "track_id" : the track id
        - "tree_id" : the tree id
    split_df : pd.DataFrame
        the splitting dataframe, with the following columns:
        - "parent_track_id" : the track id of the parent
        - "child_track_id" : the track id of the parent
    merge_df : pd.DataFrame
        the splitting dataframe, with the following columns:
        - "parent_track_id" : the track id of the parent
        - "child_track_id" : the track id of the parent
    """
    df_data = []
    node_values = np.array(list(tree.nodes))
    frames = np.unique(node_values[:, 0])
    for frame in frames:
        indices = node_values[node_values[:, 0] == frame, 1]
        df_data.append(
            pd.DataFrame(
                {
                    "frame": [frame] * len(indices),
                    "index": indices,
                }
            )
        )
    df = pd.concat(df_data).set_index(["frame", "index"])
    connected_components = list(nx.connected_components(tree))
    for track_id, nodes in enumerate(connected_components):
        for (frame, index) in nodes:
            df.loc[(frame, index), "tree_id"] = track_id
    #            tree.nodes[(frame, index)]["tree_id"] = track_id
    tree2 = tree.copy()

    splits: List[Tuple[IntTuple, List[IntTuple]]] = []
    merges: List[Tuple[IntTuple, List[IntTuple]]] = []
    for node in tree.nodes:
        frame0, _index0 = node
        neighbors = list(tree.neighbors(node))
        children = [(frame, index) for (frame, index) in neighbors if frame > frame0]
        parents = [(frame, index) for (frame, index) in neighbors if frame < frame0]
        assert len(children) + len(parents) == len(neighbors)
        if len(children) > 1:
            for child in children:
                if tree2.has_edge(node, child):
                    tree2.remove_edge(node, child)
            if node not in [p[0] for p in splits]:
                splits.append((node, children))
        if len(parents) > 1:
            for parent in parents:
                if tree2.has_edge(node, parent):
                    tree2.remove_edge(node, parent)
            if node not in [p[0] for p in merges]:
                merges.append((node, parents))

    connected_components = list(nx.connected_components(tree2))
    for track_id, nodes in enumerate(connected_components):
        for (frame, index) in nodes:
            df.loc[(frame, index), "track_id"] = track_id
    #            tree.nodes[(frame, index)]["track_id"] = track_id

    for k in ["tree_id", "track_id"]:
        df[k] = df[k].astype(int)

    split_df_data = []
    for (node, children) in splits:
        for child in children:
            split_df_data.append(
                {
                    "parent_track_id": df.loc[node, "track_id"],
                    "child_track_id": df.loc[child, "track_id"],
                }
            )
    split_df = pd.DataFrame.from_records(split_df_data).astype(int)

    merge_df_data = []
    for (node, parents) in merges:
        for parent in parents:
            merge_df_data.append(
                {
                    "parent_track_id": df.loc[parent, "track_id"],
                    "child_track_id": df.loc[node, "track_id"],
                }
            )
    merge_df = pd.DataFrame.from_records(merge_df_data).astype(int)

    return df, split_df, merge_df
